///////////////////////////////////////////////////////////////////////////////
// PjPoly class
//
// Author: Humberto Jos Bortolossi (http://www.professores.uff.br/hjbortol/)
//
// Creative Commons License: http://creativecommons.org/licenses/by-nc-sa/2.5/
//
// Date: 12/09/2010
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Native JAVA imports
///////////////////////////////////////////////////////////////////////////////
import java.awt.Color;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;


///////////////////////////////////////////////////////////////////////////////
// JavaView imports
///////////////////////////////////////////////////////////////////////////////
import jv.anim.PsAnimation;
import jv.function.PuFunction;
import jv.geom.PgElementSet;
import jv.geom.PgPointSet;
import jv.geom.PgPolygon;
import jv.geom.PgPolygonSet;
import jvx.geom.PwCleanMesh;
import jvx.geom.PwClip;
import jvx.geom.PwModel;
import jvx.geom.PwUnfold;
import jv.loader.PgJvxLoader;
import jv.number.PuDouble;
import jv.number.PuInteger;
import jv.object.PsDebug;
import jv.project.PgGeometryIf;
import jv.project.PgJvxSrc;
import jv.project.PjProject;
import jv.project.PvDisplayIf;
import jv.project.PvCameraIf;
import jv.project.PvPickListenerIf;
import jv.project.PvPickEvent;
import jv.vecmath.PdVector;
import jv.vecmath.PiVector;
import jvx.geom.PwCleanMesh;
import jvx.loader.PgOffLoader;
import jvx.surface.PgDomainDescr;
import jvx.surface.PgParmSurface;


///////////////////////////////////////////////////////////////////////////////
// Poly applet
///////////////////////////////////////////////////////////////////////////////
public class PjPoly extends PjProject implements PvPickListenerIf
{
   protected int idiom;
   protected String codebase;
   protected URL fullpath;
   protected double initial_transparency;
   protected boolean enable_choice;

   protected PuDouble transparency;
   
   protected int numberOfVertices;
   protected int numberOfEdges;
   protected int numberOfElements;

   protected String category;
   protected String solid;
   protected PgElementSet poly;

   protected double R;                    // sphere radius
   protected PgParmSurface sphere;

   protected PwModel m_pwDual;
   protected PgElementSet pseudodual;  
   protected double pseudodual_factor = 0.99;
   protected PgElementSet dual;   
   protected boolean isShowingDual = false;

   protected PgPolygon picked_point;      // Picked point
   protected PdVector rotationAxis;       // Rotation axis
   protected PuDouble translation_factor;
   protected PgElementSet cutplane;
   protected double picked_point_lon_in_deg;
   protected double picked_point_lat_in_deg;
   protected PuDouble angle_lat;   
   protected PuDouble angle_lon;   

   protected PwClip m_pwClip;
   protected String m_defClipFunc;

   protected PwUnfold m_pwUnfold;
   protected PsAnimation m_pwAnimation;
   protected PuInteger flatten_level;
   protected PvDisplayIf disp; // The PwUnfold needs a reference to the
                               // display in order to work properly.
   protected PvCameraIf cam;   

   protected PwModel m_pwModel;
   protected PuDouble m_parameter;
   protected int m_modelingType;
   protected double m_parameterTrunc;
   protected double m_parameterStellate;
   protected double m_parameterPunch;
   protected double m_truncParamLowerBound;
   protected double m_truncParamUpperBound;
   protected double m_stellateParamLowerBound;
   protected double m_stellateParamUpperBound;
   protected double m_punchParamLowerBound;
   protected double m_punchParamUpperBound;
   protected boolean trn_fill = false;
   
   protected boolean showVertices = true;
   protected boolean showEdges = true;
   protected boolean showElements = true;
   
   protected KeyListener internalKeyListener;   
   
   protected String displayTitle;
   protected String eulerCharacteristic = "Para o poliedro original:     V - A + F = ";
   protected boolean operationInUse = false;
   protected int digitsControl = 0;
   protected boolean isShowingCoordinates = false;
   protected boolean isModeling = false;
   
   protected double x_min = -9.0;
   protected double x_max = 9.0;
   protected double y_min = -9.0;
   protected double y_max = 9.0;
   protected double z_min = -9.0;
   protected double z_max = 9.0;     
   protected PgPolygonSet axes1;
   protected boolean isShowingAxes = false;
   
   protected String pdfName;
   
   protected double fieldOfView = 0.5;

   //////////////////////////////////////////////////////////////////////////
   // Default constructor
   //////////////////////////////////////////////////////////////////////////
   public PjPoly(String input_codebase,
		         int input_idiom,
		         String input_category,
		         String input_solid,
		         double input_initial_transparency,
		         boolean input_enable_choice)
   {
      super("A Plethora of Polyhedra");

      codebase = input_codebase;
      idiom = input_idiom;
      category = input_category;
      solid = input_solid;
      initial_transparency = input_initial_transparency;
      enable_choice = input_enable_choice;

      if (getClass() == PjPoly.class)
      {
         init();
      };
   }; // End of PjPoly()

   //////////////////////////////////////////////////////////////////////////
   // Initialize variables
   //////////////////////////////////////////////////////////////////////////
   public void init()
   {
       m_pwDual = new PwModel(); 
       poly = new PgElementSet(3); 
       pseudodual = new PgElementSet(3);
       dual = new PgElementSet(3);

       R = 13.0;

       // The sphere
       // Setup globe as a parametrized surface
       PuFunction m_function_c = new PuFunction(2, 3);
       m_function_c.setName("Globe");
       m_function_c.setExpression(0, Double.toString(R) + "*cos(v)*cos(u)");
       m_function_c.setExpression(1, Double.toString(R) + "*cos(v)*sin(u)");
       m_function_c.setExpression(2, Double.toString(R) + "*sin(v)");

       // Setup parametrized surface domain discretization
       PgDomainDescr m_domain_c = new PgDomainDescr(2);
       m_domain_c.setName("Globe Parametrization Domain");
       m_domain_c.setMaxSize(-Math.PI, -Math.PI/2, Math.PI, Math.PI/2);
       m_domain_c.setSize(-Math.PI, -Math.PI/2, Math.PI, Math.PI/2);
       m_domain_c.setDiscr(50, 50);
       m_domain_c.init();

       // Computes parametrized surface
       sphere = new PgParmSurface(3);
       sphere.setDomainDescr(m_domain_c);
       sphere.setFunctionExpr(m_function_c);
       sphere.compute();

       // Setup parametrized surface appearance
       sphere.setGlobalElementColor(Color.GREEN);
       sphere.showEdges(false);
       sphere.setTransparency(1.0);
       sphere.showTransparency(true);

       // Creates the picked point
       picked_point = new PgPolygon(3);
       picked_point.setName("Picked Point");
       picked_point.setNumVertices(2);
       picked_point.showVertices(true);
       picked_point.setVertexColor(1, Color.BLACK);
       picked_point.showVertexColors(true);
       picked_point.setGlobalEdgeColor(Color.BLACK);
       picked_point.setVertex(0, 0.0, 0.0, 0.0);
       double lat_in_rad = 90.0*Math.PI/180.0;
       double lon_in_rad = 90.0*Math.PI/180.0;
       double x = R*Math.cos(lat_in_rad)*Math.cos(lon_in_rad);
       double y = R*Math.cos(lat_in_rad)*Math.sin(lon_in_rad);
       double z = R*Math.sin(lat_in_rad);
       picked_point.setVertex(1, x, y, z);
       picked_point.getVertex(0).setName("");
       picked_point.getVertex(1).setName("e");
       picked_point.setLabelColor(0, Color.BLACK);
       picked_point.setLabelColor(1, Color.BLACK);
       picked_point.setLabelSize(0, 20);
       picked_point.setLabelSize(1, 20);
       picked_point.showVertexLabels(true);
       picked_point.showPolygonEndArrow(true);

       // Rotation axis
       PdVector v1 = new PdVector(0, 0, 1);
       PdVector v2 = new PdVector(picked_point.getVertex(1).getEntry(0),
                                  picked_point.getVertex(1).getEntry(1),
                                  picked_point.getVertex(1).getEntry(2));
       rotationAxis = PdVector.crossNew(v1, v2);
       rotationAxis.normalize();

       // Cut plane
       cutplane = new PgElementSet(3); // This is a 3D object
       cutplane.setNumElements(1);
       cutplane.setNumVertices(4); // Our poly has 8 vertices
       cutplane.setVertex(0,  R,  R, 0);
       cutplane.setVertex(1, -R,  R, 0);
       cutplane.setVertex(2, -R, -R, 0);
       cutplane.setVertex(3,  R, -R, 0);
       cutplane.setElement(0, 0, 1, 2, 3);
       cutplane.showElementColors(true);
       cutplane.showElements(true);
       cutplane.showEdges(true);
       cutplane.setGlobalElementColor(Color.BLUE);
       cutplane.setTransparency(0.7);
       cutplane.showTransparency(true);

       // Rotation around e-axis slider configuration
       translation_factor = new PuDouble("Translao:", this);     
       translation_factor.setDefBounds(-R, R, 0.1, 0.2);
       translation_factor.setDefValue(0.0);
       translation_factor.init();
       translation_factor.setEnabled(false);
       translation_factor.setEnabledConfigButton(false);

       // Latitude
       angle_lat = new PuDouble("ngulo 1:", this);     
       angle_lat.setDefBounds(-90.0, 90.0, 0.5, 1.0);
       angle_lat.setDefValue(90.0);
       angle_lat.init();
       angle_lat.setEnabled(false);
       angle_lat.setEnabledConfigButton(false);
       
       // Longitude
       angle_lon = new PuDouble("ngulo 2:", this);     
       angle_lon.setDefBounds(-180.0, 180.0, 0.5, 1.0);
       angle_lon.setDefValue(90.0);
       angle_lon.init();
       angle_lon.setEnabled(false);
       angle_lon.setEnabledConfigButton(false);
       
       // Flatten level slider
       flatten_level = new PuInteger("Nvel:", this);     
       flatten_level.setDefBounds(0, 100, 2, 4);
       flatten_level.setDefValue(0);
       flatten_level.init();
       flatten_level.setEnabled(false);
       flatten_level.setEnabledConfigButton(false);
       
       // Transparency slider
       transparency = new PuDouble("Transparncia:", this);     
       transparency.setDefBounds(0.0, 1.0, 0.01, 0.05);
       transparency.setDefValue(initial_transparency);
       transparency.init();
       transparency.setEnabled(true); 
       transparency.setEnabledConfigButton(false);

       // Model parameter sliders
       m_modelingType = PwModel.PUNCH;
       m_parameterTrunc = 0.3;
       m_parameterStellate = 1.0;
       m_parameterPunch = 0.78;
       m_truncParamLowerBound = 0.0;
       m_truncParamUpperBound = 0.5;
       m_stellateParamLowerBound = -6.0;
       m_stellateParamUpperBound = 6.0;
       m_punchParamLowerBound = 0.0;
       m_punchParamUpperBound = 1.0;
       m_parameter = new PuDouble("Parmetro:", this);
       m_parameter.setDefBounds(m_punchParamLowerBound,
                                m_punchParamUpperBound, 0.01, 0.1);
       m_parameter.setDefValue(m_parameterPunch);
       m_parameter.init();      
       m_parameter.setEnabled(false);
       m_parameter.setEnabledConfigButton(false);
       
       // Axes
       // Axes 1 configuration
       axes1 = new PgPolygonSet(3); // This is a 3D object
       axes1.setNumVertices(6);
       axes1.setNumPolygons(3);
       axes1.setVertex(0, new PdVector(x_min, 0.0, 0.0));
       axes1.setVertex(1, new PdVector(0.0, y_min, 0.0));
       axes1.setVertex(2, new PdVector(0.0, 0.0, z_min));
       axes1.setVertex(3, new PdVector(x_max, 0.0, 0.0));
       axes1.setVertex(4, new PdVector(0.0, y_max, 0.0));
       axes1.setVertex(5, new PdVector(0.0, 0.0, z_max));
       axes1.getVertex(0).setName("");
       axes1.getVertex(1).setName("");
       axes1.getVertex(2).setName("");
       axes1.getVertex(3).setName("x");
       axes1.getVertex(4).setName("y");
       axes1.getVertex(5).setName("z");
       axes1.showVertexLabels(true);
       axes1.setPolygon(0, 0, 3);
       axes1.setPolygon(1, 1, 4);
       axes1.setPolygon(2, 2, 5);
       axes1.setGlobalPolygonColor(Color.BLUE);
       axes1.setGlobalVertexColor(Color.BLUE);
       axes1.setPolygonSize(0, 0.5);
       axes1.setPolygonSize(1, 0.5);
       axes1.setPolygonSize(2, 0.5);
       axes1.showPolygonSizes(true);
       axes1.showPolygonEndArrow(true);       
   }; // End of init()


   //////////////////////////////////////////////////////////////////////////
   // Load off file
   //////////////////////////////////////////////////////////////////////////
   public void loadGeometryFromFile(Model object) 
   {		
		
	  String filename = object.getFileName();
	  String filetype = object.getFileType();	  

      String pathname = "/off/" + filename + "." + filetype;      
      
      {         
         
         BufferedReader in = null; 
         in = new BufferedReader(new InputStreamReader(getClass().getResourceAsStream(pathname)));
                  
         // BufferedReader in = new BufferedReader(new InputStreamReader(r));
         if(in == null)
         {
            PsDebug.warning("could not open file  " + pathname + "...");
            System.out.println("could not open file  " + pathname + "...");
         }
	     else
         {
            // Loading polyhedron
	    	PgJvxSrc[] geometries; 
	    	if (filetype.equals("off"))
	    	{
	    	   PgOffLoader offloader = new PgOffLoader(); 
               offloader.load(in);            
               geometries = offloader.getGeometries();
               poly.setJvx(geometries[0]);
               pseudodual.setJvx(geometries[0]);
	    	}
	    	if (filetype.equals("jvx"))
	    	{
	    	   PgJvxLoader jvxloader = new PgJvxLoader(); 
               jvxloader.load(in);            
               geometries = jvxloader.getGeometries();
               poly.setJvx(geometries[0]);
               pseudodual.setJvx(geometries[0]);
	    	}
	    	
	    	

	        // Configuring the polyhedron	        
            poly.setTransparency(transparency.getValue());
            poly.showVertices(showVertices);
            if (transparency.getValue() > 0.05)
            {
               poly.showTransparency(true);
            }
            else
            {
               poly.showTransparency(false);	
            }
            poly.showElementColors(true);
            poly.showElements(showElements);
            poly.showEdges(showEdges);
            poly.showVertexOutline(true);
            poly.setGlobalEdgeColor(Color.BLACK);
            poly.setGlobalVertexColor(Color.WHITE);
            poly.showVertexColors(true);

            PwCleanMesh.identifyVertices(poly, 1E-010D);
            poly.makeNeighbour();
            poly.close();
            
            // Add because key events are now internally handled
            poly.showVertexLabels(false);            
            
            // if (poly.getDiameter()/2.0 > 1.5)
            {
                R = poly.getDiameter()/2.5;
                translation_factor.setBounds(-R, R);

                // The sphere
                // Setup globe as a parametrized surface
                PuFunction m_function_c = new PuFunction(2, 3);
                m_function_c.setName("Globe");
                m_function_c.setExpression(0, Double.toString(R) + "*cos(v)*cos(u)");
                m_function_c.setExpression(1, Double.toString(R) + "*cos(v)*sin(u)");
                m_function_c.setExpression(2, Double.toString(R) + "*sin(v)");

                // Setup parametrized surface domain discretization
                PgDomainDescr m_domain_c = new PgDomainDescr(2);
                m_domain_c.setName("Globe Parametrization Domain");
                m_domain_c.setMaxSize(-Math.PI, -Math.PI/2, Math.PI, Math.PI/2);
                m_domain_c.setSize(-Math.PI, -Math.PI/2, Math.PI, Math.PI/2);
                m_domain_c.setDiscr(50, 50);
                m_domain_c.init();

                // Computes parametrized surface
                sphere = new PgParmSurface(3);
                sphere.setDomainDescr(m_domain_c);
                sphere.setFunctionExpr(m_function_c);
                sphere.compute();

                // Setup parametrized surface appearance
                sphere.setGlobalElementColor(Color.GREEN);
                sphere.showEdges(false);
                sphere.setTransparency(1.0);
                sphere.showTransparency(true);

                // Creates the picked point
                picked_point = new PgPolygon(3);
                picked_point.setName("Picked Point");
                picked_point.setNumVertices(2);
                picked_point.showVertices(true);
                picked_point.setVertexColor(1, Color.BLACK);
                picked_point.showVertexColors(true);
                picked_point.setGlobalEdgeColor(Color.BLACK);
                picked_point.setVertex(0, 0.0, 0.0, 0.0);
                double lat_in_rad = 90.0*Math.PI/180.0;
                double lon_in_rad = 90.0*Math.PI/180.0;
                double x = R*Math.cos(lat_in_rad)*Math.cos(lon_in_rad);
                double y = R*Math.cos(lat_in_rad)*Math.sin(lon_in_rad);
                double z = R*Math.sin(lat_in_rad);
                picked_point.setVertex(1, x, y, z);
                picked_point.getVertex(0).setName("");
                picked_point.getVertex(1).setName("e");
                picked_point.setLabelColor(0, Color.BLACK);
                picked_point.setLabelColor(1, Color.BLACK);
                picked_point.setLabelSize(0, 20);
                picked_point.setLabelSize(1, 20);
                picked_point.showVertexLabels(true);
                picked_point.showPolygonEndArrow(true);

                // Rotation axis
                PdVector v1 = new PdVector(0, 0, 1);
                PdVector v2 = new PdVector(picked_point.getVertex(1).getEntry(0),
                                           picked_point.getVertex(1).getEntry(1),
                                           picked_point.getVertex(1).getEntry(2));
                rotationAxis = PdVector.crossNew(v1, v2);
                rotationAxis.normalize();

                // Cut plane
                cutplane = new PgElementSet(3); // This is a 3D object
                cutplane.setNumElements(1);
                cutplane.setNumVertices(4); // Our poly has 8 vertices
                cutplane.setVertex(0,  R,  R, 0);
                cutplane.setVertex(1, -R,  R, 0);
                cutplane.setVertex(2, -R, -R, 0);
                cutplane.setVertex(3,  R, -R, 0);
                cutplane.setElement(0, 0, 1, 2, 3);
                cutplane.showElementColors(true);
                cutplane.showElements(true);
                cutplane.showEdges(true);
                cutplane.setGlobalElementColor(Color.BLUE);
                cutplane.setTransparency(0.7);
                cutplane.showTransparency(true);   
                
                PdVector[] bb = poly.getBounds();
                double sf = 1.2;
                x_max = sf*Math.max(Math.abs(bb[0].getEntry(0)), Math.abs(bb[1].getEntry(0)));
                x_min = -x_max;
                y_max = sf*Math.max(Math.abs(bb[0].getEntry(1)), Math.abs(bb[1].getEntry(1)));
                y_min = -y_max;
                z_max = sf*Math.max(Math.abs(bb[0].getEntry(2)), Math.abs(bb[1].getEntry(2)));
                z_min = -z_max;
                axes1.setVertex(0, new PdVector(x_min, 0.0, 0.0));
                axes1.setVertex(1, new PdVector(0.0, y_min, 0.0));
                axes1.setVertex(2, new PdVector(0.0, 0.0, z_min));
                axes1.setVertex(3, new PdVector(x_max, 0.0, 0.0));
                axes1.setVertex(4, new PdVector(0.0, y_max, 0.0));
                axes1.setVertex(5, new PdVector(0.0, 0.0, z_max));
                axes1.getVertex(0).setName("");
                axes1.getVertex(1).setName("");
                axes1.getVertex(2).setName("");
                axes1.getVertex(3).setName("x");
                axes1.getVertex(4).setName("y");
                axes1.getVertex(5).setName("z");
                axes1.showVertexLabels(true);
                axes1.setPolygon(0, 0, 3);
                axes1.setPolygon(1, 1, 4);
                axes1.setPolygon(2, 2, 5);
                axes1.setGlobalPolygonColor(Color.BLUE);
                axes1.setGlobalVertexColor(Color.BLUE);
                axes1.setGlobalVertexSize(0.0);
                axes1.setPolygonSize(0, 0.5);
                axes1.setPolygonSize(1, 0.5);
                axes1.setPolygonSize(2, 0.5);
                axes1.showPolygonSizes(true);
                axes1.showPolygonEndArrow(true);
                axes1.update(axes1);
            };
            

	        // Computing pseudodual	        
            PwCleanMesh.identifyVertices(pseudodual, 1E-010D);
            pseudodual.makeNeighbour();
            pseudodual.close();
	        m_pwDual.setGeometry(pseudodual);
	        PwModel.dual(pseudodual);

            // Configuring the dual
	        PdVector center = pseudodual.getCenter();
	        double xc = center.getEntry(0);
	        double yc = center.getEntry(1);
	        double zc = center.getEntry(2);

	        for (int i = 0; i < pseudodual.getNumVertices(); i++)
            {
               double xv = pseudodual.getVertex(i).getEntry(0);
               double yv = pseudodual.getVertex(i).getEntry(1);
               double zv = pseudodual.getVertex(i).getEntry(2);
               pseudodual.getVertex(i).setEntry(0, xc + pseudodual_factor*(xv - xc));
               pseudodual.getVertex(i).setEntry(1, yc + pseudodual_factor*(yv - yc));
               pseudodual.getVertex(i).setEntry(2, zc + pseudodual_factor*(zv - zc));
            };
	    
            pseudodual.setTransparency(0.5);
            pseudodual.showVertices(true);
            pseudodual.showTransparency(true);
            pseudodual.showElementColors(true);
            pseudodual.showElements(true);
            pseudodual.showEdges(true);
            pseudodual.showVertexOutline(true);
            pseudodual.showVertexOutline(true);
            pseudodual.setGlobalElementColor(Color.YELLOW);
            
            poly.update(poly);
            pseudodual.update(pseudodual);
          
            // Topological information
	        operationInUse = false;
            displayTitle = eulerCharacteristic +
                           object.getNumberOfVertices() + " - " +
                           object.getNumberOfEdges() + " + " +
                           object.getNumberOfElements() + " = " +
                           object.getEulerCharacteristic() + ".";
            
            // Reset coordinates display
            isShowingCoordinates = false;
            
            pdfName = object.getFileName() + ".pdf";

	    // Reset cut plane controls
            translation_factor.setValue(0.0);
	    angle_lat.setValue(0.0);
	    angle_lon.setValue(0.0);
         }
      }
            
      if ((object.isEnabledDual() == true))
      {
	      pathname = "/off/" + filename + "-dual." + filetype;      
	      {
	         BufferedReader in = null;
                 in = new BufferedReader(new InputStreamReader(getClass().getResourceAsStream(pathname)));


	         if(in == null)
	         {
	            PsDebug.warning("could not open file  " + pathname + "...");
	            System.out.println("could not open file  " + pathname + "...");
	         }
		     else
	         {
	            // Loading polyhedron
		    	PgJvxSrc[] geometries; 
		    	if (filetype.equals("off"))
		    	{
		    	   PgOffLoader offloader = new PgOffLoader(); 
	               offloader.load(in);            
	               geometries = offloader.getGeometries();
	               dual.setJvx(geometries[0]);
		    	}
		    	if (filetype.equals("jvx"))
		    	{
		    	   PgJvxLoader jvxloader = new PgJvxLoader(); 
	               jvxloader.load(in);            
	               geometries = jvxloader.getGeometries();
	               dual.setJvx(geometries[0]);
		    	}
		    		           	
		    
	            dual.setTransparency(0.5);
	            dual.showVertices(true);
	            dual.showTransparency(true);
	            dual.showElementColors(true);
	            dual.showElements(true);
	            dual.showEdges(true);
	            dual.showVertexOutline(true);
	            dual.setGlobalElementColor(Color.BLUE);

	            isShowingDual = true;
	            
	            dual.update(dual);

	         }
	      }
      }
      else
      {
    	  if (isShowingDual == true)
    	  {
	          isShowingDual = false;
	          
    		  removeGeometry(dual);
    		  
    	  }
      }      
   }; // End of loadGeometryFromFile() 
   
   
   //////////////////////////////////////////////////////////////////////////
   // updateCutPlane()
   //////////////////////////////////////////////////////////////////////////
   public void updateCutPlane()
   {
      boolean isShowingVertexLabels = poly.isShowingVertexLabels();
      
      // Restore original coordinates
      cutplane.setVertex(0,  R,  R, 0);
      cutplane.setVertex(1, -R,  R, 0);
      cutplane.setVertex(2, -R, -R, 0);
      cutplane.setVertex(3,  R, -R, 0);

      // e is the unit normal of the cut plane 
      PdVector e = new PdVector(picked_point.getVertex(1).getEntry(0),
                                picked_point.getVertex(1).getEntry(1),
                                picked_point.getVertex(1).getEntry(2));
      e.normalize();

      // Using quaternions to compute the inclinitation of the cut plane
      PdVector z = new PdVector(0, 0, 1);
      double a = PdVector.angle(e, z)*Math.PI/180.0;
      double ca = Math.cos(a/2.0);
      double sa = Math.sin(a/2.0);
      Quaternion qo = new Quaternion(ca, +sa*rotationAxis.getEntry(0),
                                         +sa*rotationAxis.getEntry(1),
                                         +sa*rotationAxis.getEntry(2));
      Quaternion qc = new Quaternion(ca, -sa*rotationAxis.getEntry(0),
                                         -sa*rotationAxis.getEntry(1),
                                         -sa*rotationAxis.getEntry(2));
      Quaternion r = new Quaternion();
      Quaternion w = new Quaternion();
      PdVector d = new PdVector();

      for (int i = 0; i < cutplane.getNumVertices(); i++)
      {
         w.setComponent(0.0, cutplane.getVertex(i).getEntry(0),
                             cutplane.getVertex(i).getEntry(1),
                             cutplane.getVertex(i).getEntry(2));
         r.setComponent(qo);
         r.multiply(w);
         r.multiply(qc);

         d.setEntry(0, e.getEntry(0));
         d.setEntry(1, e.getEntry(1));
         d.setEntry(2, e.getEntry(2));
         d.multScalar(translation_factor.getValue());
         d.add(r.getVectorPart());
         cutplane.setVertex(i, d);
      };
      cutplane.update(cutplane);
     
      // The cut operation is defining using clippiing by a level set of
      // a linear function
      double eu = Math.round(e.getEntry(0)*100.0)/100.0;
      double ev = Math.round(e.getEntry(1)*100.0)/100.0;
      double ew = Math.round(e.getEntry(2)*100.0)/100.0;
      if (eu > 0.0)
      {
         m_defClipFunc =  Double.toString(Math.abs(eu)) + "*u";
      }
      else
      {
         m_defClipFunc = "-" + Double.toString(Math.abs(eu)) + "*u";
      };
      if (ev > 0.0)
      {
         m_defClipFunc = 
                   m_defClipFunc + "+" + Double.toString(Math.abs(ev)) + "*v";
      }
      else
      {
         m_defClipFunc = 
                   m_defClipFunc + "-" + Double.toString(Math.abs(ev)) + "*v";
      };
      if (ew > 0.0)
      {
         m_defClipFunc = 
                   m_defClipFunc + "+" + Double.toString(Math.abs(ew)) + "*w";
      }
      else
      {
         m_defClipFunc = 
                   m_defClipFunc + "-" + Double.toString(Math.abs(ew)) + "*w";
      };
      m_pwClip.setFunction(m_defClipFunc);
      m_pwClip.setLevel(translation_factor.getValue()); 
      m_pwClip.clipAtFunction();
      
      poly.showVertexLabels(isShowingVertexLabels); // For some reason m_pwClip
                                                    // resets displaying of vertex
                                                    // labels, so we remember it.    
      poly.update(poly);
    
      updateCoordinatesLabels();
   }; // End of updateCutPlane()


   //////////////////////////////////////////////////////////////////////////
   // setModelingType()
   //////////////////////////////////////////////////////////////////////////
   public void setModelingType(int i)
   {
       if (m_modelingType == PwModel.TRUNCATEFILL)
       {
           m_truncParamUpperBound = m_parameter.getMax();
           m_truncParamLowerBound = m_parameter.getMin();
       } else
       if (m_modelingType == PwModel.TRUNCATE)
       {
           m_truncParamUpperBound = m_parameter.getMax();
           m_truncParamLowerBound = m_parameter.getMin();
       } else
       if (m_modelingType == PwModel.STELLATE)
       {
           m_stellateParamUpperBound = m_parameter.getMax();
           m_stellateParamLowerBound = m_parameter.getMin();
       } else
       if (m_modelingType == PwModel.PUNCH)
       {
           m_punchParamUpperBound = m_parameter.getMax();
           m_punchParamLowerBound = m_parameter.getMin();
       };

       if (i == PwModel.TRUNCATEFILL)
       {
           m_parameter.setBounds(m_truncParamLowerBound,
                                                      m_truncParamUpperBound);
           m_parameter.setValue(m_parameterTrunc);
       } else
       if (i == PwModel.TRUNCATE)
       {
           m_parameter.setBounds(m_truncParamLowerBound,
                                                      m_truncParamUpperBound);
           m_parameter.setValue(m_parameterTrunc);
       } else
       if (i == PwModel.STELLATE)
       {
           m_parameter.setBounds(m_stellateParamLowerBound,
                                                   m_stellateParamUpperBound);
           m_parameter.setValue(m_parameterStellate);
       } else
       if (i == PwModel.PUNCH)
       {
           m_parameter.setBounds(m_punchParamLowerBound,
                                                      m_punchParamUpperBound);
           m_parameter.setValue(m_parameterPunch);
       };

       m_modelingType = i;
   }; // End of setModelingType()


   //////////////////////////////////////////////////////////////////////////
   // Things to do when the class is started
   //////////////////////////////////////////////////////////////////////////
   public void start()
   {
      super.start();
      addGeometry(sphere); // We use the sphere to take pick points for the
                           // cut plane operation. So, it must be included
			               // first.

      // Necessary to apply PwWorkshop operations
      PwCleanMesh.identifyVertices(poly, 1E-010D);
      poly.makeNeighbour();

      addGeometry(poly);
      // addGeometry(dual);
      selectGeometry(sphere);
      removeGeometry(sphere);

      update(this);

      disp = getDisplay(); // The PwUnfold needs a reference to the
                           // display in order to work properly.
      disp.fit();
  	  cam = disp.getCamera();
      
      // Key events
      internalKeyListener = new KeyAdapter() 
      {
          public void keyPressed(KeyEvent keyEvent) 
          {
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_1)
        	  {        	  
	        	  if (poly.isShowingVertexLabels() == true)
	        	  {
	        	     poly.showVertexLabels(false);
	        	  }
	        	  else
	        	  {
	         	     poly.showVertexLabels(true);
	        		 if (isModeling == true)
	        		 {
                        if (isShowingCoordinates == true)
                        {
                           updateCoordinatesLabels();
                        }
	        		 }
	         	     
	        	  }
	              poly.update(poly);
        	  };
        	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_2)
        	  {
      			  if ((poly.isShowingVertexLabels() == true) && (isShowingCoordinates == true))
       			  {
          		     digitsControl++;
           		     digitsControl = digitsControl % 8;
       			  }
      			  else
      			  {
      				  isShowingCoordinates = true;
      			  };        		  
        		  
        		  long nd = digitsControl + 1;
        		  long p = (long) Math.pow(10, nd);        		  
        		  long d;
        		  for (int i = 0; i < poly.getNumVertices(); i++)
        		  {
        			  PdVector v = poly.getVertex(i);
        			  String x = Double.toString(Math.round(v.getEntry(0)*p)/(p*1.0));
        			  d = x.length() - x.indexOf('.');
        			  if (nd >= d)
        			  {
        				  for (int j = 0; j <= nd - d; j++)
        				  {
        					  x = x + "0";
        				  }
        			  }
        			  
        			  String y = Double.toString(Math.round(v.getEntry(1)*p)/(p*1.0));
        			  d = y.length() - y.indexOf('.');
        			  if (nd >= d)
        			  {
        				  for (int j = 0; j <= nd - d; j++)
        				  {
        					  y = y + "0";
        				  }
        			  }
        			  
        			  String z = Double.toString(Math.round(v.getEntry(2)*p)/(p*1.0));
        			  d = z.length() - z.indexOf('.');
        			  if (nd >= d)
        			  {
        				  for (int j = 0; j <= nd - d; j++)
        				  {
        					  z = z + "0";
        				  }
        			  }
        			  
        			  v.setName("(" + x + ", " + y + "," + z + ")");
        		  };      			  
        		  poly.update(poly);
        	  };
        	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_3)
        	  {
        		  isShowingCoordinates = false;
        		  for (int i = 0; i < poly.getNumVertices(); i++)
        		  {
        			  poly.getVertex(i).setName(Integer.toString(i));
        		  };
        		  poly.update(poly);
        	  };
        	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_4)
        	  {
        		  if (isShowingAxes == false)
        		  {
        			  disp.addGeometry(axes1);
        			  isShowingAxes = true;
        		  }
        		  else
        		  {
        			  disp.removeGeometry(axes1);
        			  isShowingAxes = false;        			  
        		  };
        		  disp.update(disp);
        	  };
        	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_5)
        	  {
        		  if (disp.isShowingBndBox() == false)
        		  {
        			  disp.showBndBox(true);
        		  }
        		  else
        		  {
        			  disp.showBndBox(false);        			  
        		  };
        	  };
        	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_6)
        	  {
        		  fieldOfView -= 0.05;
        		  if (fieldOfView <= 0.0)
        		  {
        			  fieldOfView = 0.0;
        		  };
        		  cam.setFieldOfView(fieldOfView);
        		  cam.update(cam);
        	  };        	  
        	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_7)
        	  {
        		  fieldOfView += 0.05;
        		  if (fieldOfView >= 2.0)
        		  {
        			  fieldOfView = 2.0;
        		  };
        		  cam.setFieldOfView(fieldOfView);
        		  cam.update(cam);
        	  };        	  
        	          	          	  
        	  if (keyEvent.getKeyCode() == KeyEvent.VK_9)
        	  {        	  
        		  if (operationInUse == true)
        		  {        		  
 	        	     disp.showTitle(false);
        		  }
        		  else
        		  {
		        	  if (disp.isShowingTitle() == true)
		        	  {
		        	     disp.showTitle(false);
		        	  }
		        	  else
		        	  {
		        		 disp.setName(displayTitle);
		        		 disp.showTitle(true);        		  
		        	  }
		        	  disp.update(disp);        			  
        		  }
        	  };
        	  
        	  if ((keyEvent.getKeyCode() == KeyEvent.VK_0) || (keyEvent.getKeyCode() == KeyEvent.VK_8))
        	  {        	  
        		 fieldOfView = 0.5;
                 cam.setFieldOfView(fieldOfView);
                 cam.update(cam);
                 disp.fit();
        	  };        	  
          }
       };
       disp.addKeyListener(internalKeyListener);            
   }; // End of start()


   /////////////////////////////////////////////////////////////////////////
   // Update the class whenever a child has changed. Method is usually
   // invoked from the children.
   /////////////////////////////////////////////////////////////////////////
   public boolean update(Object event)
   {
      if (event == null)
      {
         return(true);
      };
      
      // Solid transparency
      if (event == transparency)
      {
          poly.setTransparency(transparency.getValue());

          if (transparency.getValue() > 0.05)
          {
             poly.showTransparency(true);
          }
          else
          {
             poly.showTransparency(false);	
          }
          
          poly.update(poly);
          
          return(true);
      };
      

      // Cut plane translation factor
      if (event == translation_factor)
      {
         updateCutPlane();

         return(true);
      };
      
      // Cut plane translation factor
      if ((event == angle_lat) || (event == angle_lon))
      {
    	 double lat_in_rad = angle_lat.getValue()*Math.PI/180.0;
   	     double lon_in_rad = angle_lon.getValue()*Math.PI/180.0;    	  
         double x = R*Math.cos(lat_in_rad)*Math.cos(lon_in_rad);
         double y = R*Math.cos(lat_in_rad)*Math.sin(lon_in_rad);
         double z = R*Math.sin(lat_in_rad);         
         double norm = Math.sqrt(x*x + y*y + z*z);
         if (norm == 0.0)
         {
            double l_lat_in_rad = 45.0*Math.PI/180.0;
            double l_lon_in_rad = 90.0*Math.PI/180.0;
            x = R*Math.cos(l_lat_in_rad)*Math.cos(l_lon_in_rad);
            y = R*Math.cos(l_lat_in_rad)*Math.sin(l_lon_in_rad);
            z = R*Math.sin(l_lat_in_rad);
            rotationAxis.setEntry(0, x);
            rotationAxis.setEntry(1, y);
            rotationAxis.setEntry(2, z);

            PdVector v1 = new PdVector(0, 0, 1);
            PdVector v2 = new PdVector(x, y, z);
            rotationAxis.cross(v1, v2);
            rotationAxis.normalize();
         }
         else
         {
            x /= norm;
            y /= norm;
            z /= norm;
            rotationAxis.setEntry(0, x);
            rotationAxis.setEntry(1, y);
            rotationAxis.setEntry(2, z);

            PdVector v1 = new PdVector(0, 0, 1);
            PdVector v2 = new PdVector(x, y, z);
            rotationAxis.cross(v1, v2);
            rotationAxis.normalize();

            x *= R;
            y *= R;
            z *= R;
         };

         picked_point.setVertex(1, x, y, z);
         picked_point.update(picked_point);
         
         updateCutPlane();

         return(true);
      };
      

      // Unfold level
      if (event == flatten_level)
      {
         m_pwUnfold.flatten(flatten_level.getValue()/100.0);                  
         poly.update(poly);
         
         updateCoordinatesLabels();         
         
         return(true);
      };

      // Model operation
      if (event == m_parameter)
      {
    	 boolean isShowingVertexLabels = poly.isShowingVertexLabels();
    	 
         if ((m_modelingType == PwModel.TRUNCATE) ||
             (m_modelingType == PwModel.TRUNCATEFILL))
         {
            m_parameterTrunc = m_parameter.getValue();
            poly.copy(m_pwModel.getSavedGeometry()); 
            PwModel.truncate(poly, m_parameterTrunc, trn_fill);
         };

         if (m_modelingType == PwModel.PUNCH)
         {
            m_parameterPunch = m_parameter.getValue();
            m_pwModel.setModelingType(PwModel.PUNCH);
            poly.copy(m_pwModel.getSavedGeometry()); 
            PwModel.punch(poly, m_parameterPunch);
         };

         if (m_modelingType == PwModel.STELLATE)
         {
            m_parameterStellate = m_parameter.getValue();
            m_pwModel.setModelingType(PwModel.STELLATE);
            poly.copy(m_pwModel.getSavedGeometry()); 
            PwModel.stellate(poly, m_parameterStellate);
         };

         poly.showVertexLabels(isShowingVertexLabels); // For some reason m_pwClip
                                                       // resets displaying of vertex
                                                       // labels, so we remember it.             
         
         poly.update(poly);
                  
         updateCoordinatesLabels();

         return(true);
      };

      return(super.update(event));
   }; // End of update()
   
   protected void updateCoordinatesLabels()
   {
       if ((isShowingCoordinates == true) && (poly.isShowingVertexLabels() == true))
 	   {
   		  long nd = digitsControl + 1;
		  long p = (long) Math.pow(10, nd);        		  
		  long d;
		  for (int i = 0; i < poly.getNumVertices(); i++)
		  {
			  PdVector v = poly.getVertex(i);
			  String x = Double.toString(Math.round(v.getEntry(0)*p)/(p*1.0));
			  d = x.length() - x.indexOf('.');
			  if (nd >= d)
			  {
				  for (int j = 0; j <= nd - d; j++)
				  {
					  x = x + "0";
				  }
			  }
			  
			  String y = Double.toString(Math.round(v.getEntry(1)*p)/(p*1.0));
			  d = y.length() - y.indexOf('.');
			  if (nd >= d)
			  {
				  for (int j = 0; j <= nd - d; j++)
				  {
					  y = y + "0";
				  }
			  }
			  
			  String z = Double.toString(Math.round(v.getEntry(2)*p)/(p*1.0));
			  d = z.length() - z.indexOf('.');
			  if (nd >= d)
			  {
				  for (int j = 0; j <= nd - d; j++)
				  {
					  z = z + "0";
				  }
			  }
			  
			  v.setName("(" + x + ", " + y + "," + z + ")");
		   };
		   
		   poly.update(poly);
	   };	   
   }; // End of updateCoordinateLabels()

   
   /***************************************************************************
    ***************************************************************************
    **  Begin of PvPickListenerIf Implementation
    ***************************************************************************
    ***************************************************************************/

    //////////////////////////////////////////////////////////////////////////
    // The name of a listeners allows the display to issue verbal debug
    // messages.
    //////////////////////////////////////////////////////////////////////////
    public String getName()
    {
       return("EarthPicker");
    }; // End of getName()

    //////////////////////////////////////////////////////////////////////////
    // Get a location in the display with 2d display and 3d world coordinates.
    // Method is called when display is in mode PvDisplayIf.MODE_DISPLAY_PICK.
    //////////////////////////////////////////////////////////////////////////
    public void pickDisplay(PvPickEvent pos)
    {
       // PsDebug.message("pickDisplay() called ...");
       // System.out.println("pickDisplay() called ...");
    }; // End of pickDisplay()

    //////////////////////////////////////////////////////////////////////////
    // Drag a location in the display with 2d display and 3d world coordinates.
    // Method is called when display is in mode PvDisplayIf.MODE_DISPLAY_PICK.
    //////////////////////////////////////////////////////////////////////////
    public void dragDisplay(PvPickEvent pos)
    {
       // PsDebug.message("pickDisplay() called ...");
       // System.out.println("pickDisplay() called ...");
    }; // End of dragDisplay()

    //////////////////////////////////////////////////////////////////////////
    // Pick an arbitrary point on a geometry, point may lie inside an element.
    // Method is called when display is in mode PvDisplayIf.MODE_INITIAL_PICK
    // or if temporarily the i-key is pressed, and any pixel in the display is
    // picked.
    //////////////////////////////////////////////////////////////////////////
    public void pickInitial(PvPickEvent pos)
    {
       // PsDebug.message("pickInitial() called ...");
       // System.out.println("pickInitial() called ...");
       double x = pos.getVertex().getEntry(0);
       double y = pos.getVertex().getEntry(1);
       double z = pos.getVertex().getEntry(2);
       double norm = Math.sqrt(x*x + y*y + z*z);
       if (norm == 0.0)
       {
          double lat_in_rad = 45.0*Math.PI/180.0;
          double lon_in_rad = 90.0*Math.PI/180.0;
          x = R*Math.cos(lat_in_rad)*Math.cos(lon_in_rad);
          y = R*Math.cos(lat_in_rad)*Math.sin(lon_in_rad);
          z = R*Math.sin(lat_in_rad);
          rotationAxis.setEntry(0, x);
          rotationAxis.setEntry(1, y);
          rotationAxis.setEntry(2, z);

          PdVector v1 = new PdVector(0, 0, 1);
          PdVector v2 = new PdVector(x, y, z);
          rotationAxis.cross(v1, v2);
          rotationAxis.normalize();
       }
       else
       {
          x /= norm;
          y /= norm;
          z /= norm;
          rotationAxis.setEntry(0, x);
          rotationAxis.setEntry(1, y);
          rotationAxis.setEntry(2, z);

          PdVector v1 = new PdVector(0, 0, 1);
          PdVector v2 = new PdVector(x, y, z);
          rotationAxis.cross(v1, v2);
          rotationAxis.normalize();

          x *= R;
          y *= R;
          z *= R;
       };

       picked_point.setVertex(1, x, y, z);
       picked_point.update(picked_point);
       updateCutPlane();
       
       double lat_in_deg;
       double lon_in_deg;

       if ((x <= 0.0) && (y == 0.0))
       {
          lon_in_deg = 180.0;
          lat_in_deg = Math.atan2(z, Math.abs(x))*180.0/Math.PI;
       }
       else
       {
          lat_in_deg = Math.atan2(z, Math.sqrt(x*x + y*y))*180.0/Math.PI;
          lon_in_deg = Math.atan2(y, x)*180/Math.PI;
       };

       angle_lat.setValue(lat_in_deg);
       angle_lon.setValue(lon_in_deg);       
    }; // End of pickInitial()

    //////////////////////////////////////////////////////////////////////////
    // Drag an arbitrary point along a geometry, point may lie inside an
    // element. Method is called when display is in mode
    // PvDisplayIf.MODE_INITIAL_PICK or if temporarily the i-key is pressed,
    // and any pixel in the display is dragged.
    //////////////////////////////////////////////////////////////////////////
    public void dragInitial(PvPickEvent pos)
    {
       // PsDebug.message("dragInitial() called ...");
       // System.out.println("dragInitial() called ...");
       double x = pos.getVertex().getEntry(0);
       double y = pos.getVertex().getEntry(1);
       double z = pos.getVertex().getEntry(2);
       double norm = Math.sqrt(x*x + y*y + z*z);
       if (norm == 0.0)
       {
          double lat_in_rad = 45.0*Math.PI/180.0;
          double lon_in_rad = 90.0*Math.PI/180.0;
          x = R*Math.cos(lat_in_rad)*Math.cos(lon_in_rad);
          y = R*Math.cos(lat_in_rad)*Math.sin(lon_in_rad);
          z = R*Math.sin(lat_in_rad);
          rotationAxis.setEntry(0, x);
          rotationAxis.setEntry(1, y);
          rotationAxis.setEntry(2, z);

          PdVector v1 = new PdVector(0, 0, 1);
          PdVector v2 = new PdVector(x, y, z);
          rotationAxis.cross(v1, v2);
          rotationAxis.normalize();
       }
       else
       {
          x /= norm;
          y /= norm;
          z /= norm;
          rotationAxis.setEntry(0, x);
          rotationAxis.setEntry(1, y);
          rotationAxis.setEntry(2, z);

          PdVector v1 = new PdVector(0, 0, 1);
          PdVector v2 = new PdVector(x, y, z);
          rotationAxis.cross(v1, v2);
          rotationAxis.normalize();

          x *= R;
          y *= R;
          z *= R;
       };

       picked_point.setVertex(1, x, y, z);
       picked_point.update(picked_point);
       updateCutPlane();
       
       double lat_in_deg;
       double lon_in_deg;

       if ((x <= 0.0) && (y == 0.0))
       {
          lon_in_deg = 180.0;
          lat_in_deg = Math.atan2(z, Math.abs(x))*180.0/Math.PI;
       }
       else
       {
          lat_in_deg = Math.atan2(z, Math.sqrt(x*x + y*y))*180.0/Math.PI;
          lon_in_deg = Math.atan2(y, x)*180/Math.PI;
       };

       angle_lat.setValue(lat_in_deg);
       angle_lon.setValue(lon_in_deg);       
    }; // End of dragInitial()

    //////////////////////////////////////////////////////////////////////////
    // Get a picked vertex of a geometry.
    // Method is called when display is in mode PvDisplayIf.MODE_PICK
    // or if temporarily the p-key is pressed, and a vertex picked.
    //////////////////////////////////////////////////////////////////////////
    public void pickVertex(PgGeometryIf geom, int index, PdVector vertex)
    {
        // PsDebug.message("pickVertex() called ...");
        // System.out.println("pickVertex() called ...");
    }; // End of dragInitial();

    //////////////////////////////////////////////////////////////////////////
    // Method is called when display is in mode PvDisplayIf.MODE_PICK
    // or if temporarily the p-key is pressed, and a vertex dragged.
    //////////////////////////////////////////////////////////////////////////
    public void dragVertex(PgGeometryIf a_geom, int index, PdVector vertex)
    {
       // PsDebug.message("dragVertex() called ...");
       // System.out.println("dragVertex() called ...");
    }; // End of dragVertex()

    //////////////////////////////////////////////////////////////////////////
    // Mark a set of vertices of a geometry within a given bounding box.
    // Method is called when display is in mode PvDisplayIf.MODE_MARK
    // or if temporarily the m-key is pressed, and a rectangle is drawn.
    //////////////////////////////////////////////////////////////////////////
    public void markVertices(PvPickEvent markBox)
    {
       // PsDebug.message("markVertices() called ...");
       // System.out.println("markVertices() called ...");
    }; // End of markVertices()

    //////////////////////////////////////////////////////////////////////////
    // Unmark a set of vertices of a geometry within a given bounding box.
    // Method is called when display is in mode PvDisplayIf.MODE_UNMARK
    // or if temporarily the u-key is pressed, and a rectangle is drawn.
    //////////////////////////////////////////////////////////////////////////
    public void unmarkVertices(PvPickEvent markBox)
    {
       // PsDebug.message("unmarkVertices() called ...");
       // System.out.println("unmarkVertices() called ...");
    }; // End of unmarkVertices()

   /***************************************************************************
    ***************************************************************************
    **  End of PvPickListenerIf Implementation
    ***************************************************************************
    ***************************************************************************/

}; // End of PjPoly classs
